import morphyne as mp
import time
import params_template

instance = mp.create_from_yaml(params_template.params_yaml, 0)
instance.set_non_coherent_stimulation_rate(5 / 800)

synaptic_transmission_count = 0
spike_count = 0
t_stop = 50000
batch_size = 1000

t_start = time.time()

while instance.get_t() < t_stop:
    tick_result = instance.tick_until(
        min(t_stop, instance.get_t() + batch_size))
    synaptic_transmission_count += tick_result.synaptic_transmission_count
    spike_count += len(tick_result.neuron_spikes)

t_end = time.time()

synaptic_transmission_processing_throughput = synaptic_transmission_count / \
    (t_end - t_start)

print("Spikes per cycle: {}".format(spike_count / t_stop))
print("Synaptic transmission processing throughput: {:.3e}, ({:.2f} ns per transmission)".format(
    synaptic_transmission_processing_throughput, 1e9 / synaptic_transmission_processing_throughput))



import morphyne as mp
import pandas as pd
import matplotlib.pyplot as plt
import json

params = mp.get_example_params()
print(json.dumps(params, indent=4))

instance = mp.create(params)
instance.set_non_coherent_stimulation_rate(0.0005)
instance.set_reward_rate(0.1)

stimulus_df = pd.DataFrame({"t": [0, 5, 10], "in_channel_id": [0, 1, 2]})
stimulus = mp.create_stimulus(in_channel_spikes=stimulus_df)

instance.apply_stimulus(stimulus)
tick_result = instance.tick_until(1000)

plt.scatter(tick_result.neuron_spikes.t, tick_result.neuron_spikes.nid, s=1)
plt.xlabel("t")
plt.ylabel("neuron ID")
plt.show()



params_yaml = """
layers:
- num_neurons: 800
  neuron_params:
    tau_membrane: 10.0
    refractory_period: 10
    reset_voltage: 0.0
    t_cutoff_coincidence: 20
    adaptation_threshold: 1.0
    tau_threshold: 50.0
    voltage_floor: 0.0
  plasticity_modulation_params:
    tau_eligibility_trace: 1000.0
    eligibility_trace_delay: 20
    dopamine_modulation_factor: 1.5
    t_cutoff_eligibility_trace: 1500
    dopamine_flush_period: 250
    dopamine_conflation_period: 250
- num_neurons: 200
  neuron_params:
    tau_membrane: 4.0
    refractory_period: 5
    reset_voltage: 0.0
    t_cutoff_coincidence: 20
    adaptation_threshold: 1.0
    tau_threshold: 50.0
    voltage_floor: 0.0
  plasticity_modulation_params: null
layer_connections:
- from_layer_id: 0
  to_layer_id: 0
  projection_params:
    synapse_params:
      max_weight: 0.5
      weight_scale_factor: 1.0
    stp_params: !Depression
      tau: 800.0
      p0: 0.9
      factor: 0.1
    long_term_stdp_params:
      factor_pre_before_post: 0.1
      tau_pre_before_post: 20.0
      factor_pre_after_post: -0.12
      tau_pre_after_post: 20.0
    short_term_stdp_params:
      stdp_params:
        factor_pre_before_post: 0.01
        tau_pre_before_post: 20.0
        factor_pre_after_post: 0.012
        tau_pre_after_post: 20.0
      tau: 500.0
  connect_density: 0.1
  connect_width: 2.0
  initial_syn_weight: !Randomized 0.5
  conduction_delay_max_random_part: 20
  conduction_delay_position_distance_scale_factor: 0.0
  conduction_delay_add_on: 0
  allow_self_innervation: true
- from_layer_id: 0
  to_layer_id: 1
  projection_params:
    synapse_params:
      max_weight: 0.5
      weight_scale_factor: 2.0
    stp_params: !Depression
      tau: 800.0
      p0: 0.9
      factor: 0.1
    long_term_stdp_params:
      factor_pre_before_post: 0.1
      tau_pre_before_post: 20.0
      factor_pre_after_post: -0.12
      tau_pre_after_post: 20.0
    short_term_stdp_params:
      stdp_params:
        factor_pre_before_post: 0.01
        tau_pre_before_post: 20.0
        factor_pre_after_post: 0.012
        tau_pre_after_post: 20.0
      tau: 500.0
  connect_density: 0.25
  connect_width: 2.0
  initial_syn_weight: !Randomized 0.5
  conduction_delay_max_random_part: 20
  conduction_delay_position_distance_scale_factor: 0.0
  conduction_delay_add_on: 0
  allow_self_innervation: true
- from_layer_id: 1
  to_layer_id: 0
  projection_params:
    synapse_params:
      max_weight: 0.5
      weight_scale_factor: -1.0
    stp_params: NoStp
    long_term_stdp_params: null
    short_term_stdp_params: null
  connect_density: 0.25
  connect_width: 2.0
  initial_syn_weight: !Constant 0.85
  conduction_delay_max_random_part: 0
  conduction_delay_position_distance_scale_factor: 0.0
  conduction_delay_add_on: 0
  allow_self_innervation: true
- from_layer_id: 1
  to_layer_id: 1
  projection_params:
    synapse_params:
      max_weight: 0.5
      weight_scale_factor: -1.0
    stp_params: NoStp
    long_term_stdp_params: null
    short_term_stdp_params: null
  connect_density: 0.25
  connect_width: 2.0
  initial_syn_weight: !Constant 0.85
  conduction_delay_max_random_part: 0
  conduction_delay_position_distance_scale_factor: 0.0
  conduction_delay_add_on: 0
  allow_self_innervation: true
technical_params:
  num_threads: 1
  pin_threads: false
  batched_ring_buffer_size: 21
"""



from .morphyne import *
from morphyne.instance import create, create_from_yaml, create_from_json, create_stimulus, extract_in_channel_spikes, \
    extract_force_out_channel_spikes, extract_force_neuron_spikes, Instance, Stimulus
from morphyne.example_params import get_example_params

__all__ = ["create", "create_from_yaml", "create_from_json",
           "create_stimulus", "extract_in_channel_spikes",
           "extract_force_out_channel_spikes", "extract_force_neuron_spikes",
           "Stimulus", "Instance", "get_example_params"]



N = 10000

num_I_neurons = int(N * 0.2)
num_E_neurons = N - num_I_neurons

conn_EX = {
    "from_layer_id": 0,
    "conduction_delay_position_distance_scale_factor": 0,
    "projection_params": {
        "synapse_params": {
            "weight_scale_factor": 1,
            "max_weight": 0.5
        },
        "stp_params": "NoStp",
        "long_term_stdp_params": {
            "factor_pre_before_post": 0.1,
            "tau_pre_before_post": 20,
            "factor_pre_after_post": -0.12,
            "tau_pre_after_post": 20,
        }
    },
    "connect_density": 0.01,
    "connect_width": 2,
    "conduction_delay_max_random_part": 10,
    "conduction_delay_add_on": 0,
    "allow_self_innervation": False,
    "initial_syn_weight": {
        "Constant": 0.6
    }
}

conn_IX = {
    "from_layer_id": 1,
    "conduction_delay_position_distance_scale_factor": 0,
    "projection_params": {
        "synapse_params": {
            "weight_scale_factor": -1,
            "max_weight": 10.0
        },
        "stp_params": "NoStp"
    },
    "connect_density": 0.1,
    "connect_width": 2,
    "conduction_delay_max_random_part": 0,
    "conduction_delay_add_on": 0,
    "allow_self_innervation": False,
    "initial_syn_weight": {
        "Constant": 2.0
    }
}

plasticity_modulation_params = {
    "tau_eligibility_trace": 2,
    "eligibility_trace_delay": 0,
    "t_cutoff_eligibility_trace": 2,
    "dopamine_flush_period": 1,
    "dopamine_conflation_period": 1,
    "dopamine_modulation_factor": 1.0
}

# plasticity_modulation_params = None

example_params = {
    "layers": [
        {
            "num_neurons": num_E_neurons,
            "neuron_params": {
                "tau_membrane": 20,
                "refractory_period": 10,
                "reset_voltage": 0,
                "t_cutoff_coincidence": 25,
                "adaptation_threshold": 1,
                "tau_threshold": 50,
                "voltage_floor": -10
            },
            "plasticity_modulation_params": plasticity_modulation_params,
        },
        {
            "num_neurons": num_I_neurons,
            "neuron_params": {
                "tau_membrane": 5,
                "refractory_period": 1,
                "reset_voltage": 0,
                "t_cutoff_coincidence": 1,
                "adaptation_threshold": 1,
                "tau_threshold": 50,
                "voltage_floor": 0
            },
            "plasticity_modulation_params": plasticity_modulation_params,
        },
    ],
    "layer_connections": [
        conn_EX | {"to_layer_id": 0},
        conn_EX | {"to_layer_id": 1},
        conn_IX | {"to_layer_id": 0},
        conn_IX | {"to_layer_id": 1},

    ],
    "technical_params": {
        "num_threads": 10,
        "pin_threads": False,
        "seed_override": 0,
    }
}


def get_example_params() -> dict:
    return example_params



import morphyne as mp
import params_template
import unittest


class ChecksumTest(unittest.TestCase):
    def test_vs_inner(self):
        instance = mp.create_from_yaml(params_template.params_yaml, seed=0)
        instance.set_non_coherent_stimulation_rate(0.01)
        instance.set_reward_rate(0.002)
        t_stop = 1000
        tick_result = instance.tick_until(t_stop)

        neuron_spike_checksum = tick_result.neuron_spikes.product(axis=1).sum()
        channel_spike_checksum = tick_result.out_channel_spikes.product(
            axis=1).sum()

        self.assertEqual(neuron_spike_checksum, 9325888458)
        self.assertEqual(channel_spike_checksum, 857971287)
        self.assertEqual(tick_result.synaptic_transmission_count, 5456365)

        instance.set_non_coherent_stimulation_rate(0.0)
        tick_result = instance.tick(extract_state_snapshot=True)

        self.check_single_tick_result(tick_result)

    def test_ignore_output(self):
        instance = mp.create_from_yaml(params_template.params_yaml, seed=0)
        instance.set_non_coherent_stimulation_rate(0.01)
        instance.set_reward_rate(0.002)
        t_stop = 1000
        instance.tick_until(t_stop, ignore_output=True)
        instance.set_non_coherent_stimulation_rate(0.0)
        tick_result = instance.tick(extract_state_snapshot=True)
        self.check_single_tick_result(tick_result)

    def check_single_tick_result(self, tick_result):
        voltage_checksum = tick_result.state_snapshot.membrane_voltages.sum()
        synapse_state_checksum = tick_result.state_snapshot.synapse_states.product(
            axis=1).sum()

        self.assertEqual(tick_result.neuron_spikes.nid.sum(), 39115)
        self.assertEqual(
            tick_result.out_channel_spikes.out_channel_id.sum(), 4715)
        self.assertEqual(tick_result.synaptic_transmission_count, 1525)
        self.assertAlmostEqual(voltage_checksum, 79.48388550709933)
        self.assertAlmostEqual(synapse_state_checksum,
                               94369097433.76978, places=2)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd
from .morphyne import Stimulus
from .morphyne import create_from_yaml as create_from_yaml_inner
from .morphyne import create_from_json as create_from_json_inner
import json


class StateSnapshot:

    def __init__(self, membrane_voltages: np.ndarray, synapse_states: pd.DataFrame) -> None:
        self.membrane_voltages = membrane_voltages
        self.synapse_states = synapse_states


class TickResult:

    def __init__(self, out_channel_spikes: pd.DataFrame, neuron_spikes: pd.DataFrame, state_snapshot: StateSnapshot | None, synaptic_transmission_count: int) -> None:
        self.out_channel_spikes = out_channel_spikes
        self.neuron_spikes = neuron_spikes
        self.state_snapshot = state_snapshot
        self.synaptic_transmission_count = synaptic_transmission_count


def create_from_yaml(params_yaml: str, seed: int):
    if seed is None:
        seed = 0
    return Instance(create_from_yaml_inner(params_yaml, seed))


def create_from_json(params_json: str, seed=None):
    if seed is None:
        seed = 0
    return Instance(create_from_json_inner(params_json, seed))


def create(params: dict, seed=None):
    return create_from_json(json.dumps(params), seed)


def create_stimulus(in_channel_spikes: pd.DataFrame | None = None, force_out_channel_spikes: pd.DataFrame | None = None, force_neuron_spikes: pd.DataFrame | None = None) -> Stimulus:
    stimulus = Stimulus()

    if in_channel_spikes is not None:
        stimulus.in_channel_spikes_ts = in_channel_spikes.t.to_list()
        stimulus.in_channel_spikes_ids = in_channel_spikes.in_channel_id.to_list()

    if force_out_channel_spikes is not None:
        stimulus.force_out_channel_spikes_ts = force_out_channel_spikes.t.to_list()
        stimulus.force_out_channel_spikes_ids = force_out_channel_spikes.out_channel_id.to_list()

    if force_neuron_spikes is not None:
        stimulus.force_neuron_spikes_ts = force_neuron_spikes.t.to_list()
        stimulus.force_neuron_spikes_ids = force_neuron_spikes.nid.to_list()

    return stimulus


def extract_in_channel_spikes(stimulus: Stimulus) -> pd.DataFrame:
    return pd.DataFrame({"t": stimulus.in_channel_spikes_ts, "in_channel_id": stimulus.in_channel_spikes_ids})


def extract_force_out_channel_spikes(stimulus: Stimulus) -> pd.DataFrame:
    return pd.DataFrame({"t": stimulus.force_out_channel_spikes_ts, "out_channel_id": stimulus.force_out_channel_spikes_ids})


def extract_force_neuron_spikes(stimulus: Stimulus) -> pd.DataFrame:
    return pd.DataFrame({"t": stimulus.force_neuron_spikes_ts, "nid": stimulus.force_neuron_spikes_ids})


class Instance:

    def __init__(self, inner) -> None:
        self._inner = inner

    def apply_stimulus(self, stimulus: Stimulus):
        self._inner.apply_stimulus(stimulus)

    def tick(self, spiking_in_channel_ids=[], force_spiking_out_channel_ids=[], force_spiking_nids=[], reward=None, extract_state_snapshot=False, append_to: TickResult | None = None) -> TickResult:

        if reward is None:
            reward = self._inner.reward_rate

        inner_result = self._inner.tick(
            spiking_in_channel_ids, force_spiking_out_channel_ids, force_spiking_nids, reward, extract_state_snapshot)

        spiking_out_channel_ids = np.array(
            inner_result.spiking_out_channel_ids)
        spiking_nids = np.array(inner_result.spiking_nids)

        neuron_spikes_data = {"t": inner_result.t, "nid": spiking_nids}
        out_channel_spikes_data = {"t": inner_result.t,
                                   "out_channel_id": spiking_out_channel_ids}

        df_neuron_spikes = pd.DataFrame(neuron_spikes_data, dtype=np.int64)
        df_out_channel_spikes = pd.DataFrame(
            out_channel_spikes_data, dtype=np.int64)

        state_snapshot = None
        if extract_state_snapshot:
            membrane_voltages = np.array(
                inner_result.state_snapshot.membrane_voltages)
            synapse_state_data = {"pre_syn_nid": inner_result.state_snapshot.pre_syn_nids,
                                  "post_syn_nid": inner_result.state_snapshot.post_syn_nids,
                                  "conduction_delay": inner_result.state_snapshot.conduction_delays,
                                  "weight": inner_result.state_snapshot.weights}
            df_synapse_states = pd.DataFrame(synapse_state_data)
            state_snapshot = StateSnapshot(
                membrane_voltages, df_synapse_states)

        if append_to:
            return concat_results(append_to, df_out_channel_spikes, df_neuron_spikes, state_snapshot, inner_result.synaptic_transmission_count)
        else:
            return TickResult(df_out_channel_spikes, df_neuron_spikes,
                              state_snapshot, inner_result.synaptic_transmission_count)

    def tick_until(self, t: int, ignore_output=False, append_to: TickResult | None = None) -> TickResult | None:
        if ignore_output:
            self._inner.tick_until(t)
            return None
        else:
            inner_result = self._inner.tick_until(t)
            out_channel_spikes_data = {"t": inner_result.out_channel_spikes_ts,
                                       "out_channel_id": inner_result.out_channel_spikes_ids}
            neuron_spikes_data = {
                "t": inner_result.neuron_spikes_ts, "nid": inner_result.neuron_spikes_ids}

            df_out_channel_spikes = pd.DataFrame(
                out_channel_spikes_data, dtype=np.int64)
            df_neuron_spikes = pd.DataFrame(neuron_spikes_data, dtype=np.int64)

            if append_to:
                return concat_results(append_to, df_out_channel_spikes, df_neuron_spikes, None, inner_result.synaptic_transmission_count)
            else:
                return TickResult(df_out_channel_spikes, df_neuron_spikes, None, inner_result.synaptic_transmission_count)

    def set_reward_rate(self, reward_rate):
        self._inner.reward_rate = reward_rate

    def set_non_coherent_stimulation_rate(self, rate):
        self._inner.non_coherent_stimulation_rate = rate

    def get_t(self) -> int:
        return self._inner.get_t()


def concat_results(result: TickResult, df_out_channel_spikes: pd.DataFrame, df_neuron_spikes: pd.DataFrame, state_snapshot, syn_transmission_count) -> TickResult:
    df_out_channel_spikes = pd.concat(
        [result.out_channel_spikes, df_out_channel_spikes])
    df_out_channel_spikes.reset_index(inplace=True, drop=True)

    df_neuron_spikes = pd.concat(
        [result.neuron_spikes, df_neuron_spikes])
    df_neuron_spikes.reset_index(inplace=True, drop=True)

    return TickResult(df_out_channel_spikes, df_neuron_spikes,
                      state_snapshot, syn_transmission_count)
